#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "jpeglib.h"

#define CACHE_LINE_FACTOR 16

double sinc(double x)
{
	// This is bad for very small x, should use power series instead.
	if (x == 0.0)
		return 1.0;
	else
		return sin(x) / x;
}

double lanczos_tap(double x)
{
	if (x < -3.0 || x > 3.0)
		return 0.0;
	if (x < 0.0)
		return sinc(-x*M_PI) * sinc(-x*M_PI / 3.0);
	else
		return sinc(x*M_PI) * sinc(x*M_PI / 3.0);
}


struct pix_desc {
	unsigned start, end;
	unsigned startcoeff;
};

void hscale(float *pix, unsigned char *npix, unsigned w, unsigned h, unsigned nw, unsigned dstride)
{
	struct pix_desc *pd = (struct pix_desc *)malloc(nw * sizeof(struct pix_desc));
	int size_coeffs = 8;
	float *coeffs = (float *)malloc(size_coeffs * sizeof(float));
	int num_coeffs = 0;
	int x, y, sx;
	double sf = (double)w / (double)nw;
	double support = (w > nw) ? (3.0 * sf) : (3.0 / sf);

	/* calculate the filter */
	for (x = 0; x < nw; ++x) {
		int start = ceil(x * sf - support);
		int end = floor(x * sf + support);
		double sum = 0.0;

		if (start < 0) {
			start = 0;
		}
		if (end > w - 1) {
			end = w - 1;
		}

		pd[x].start = start;
		pd[x].end = end;
		pd[x].startcoeff = num_coeffs;

		for (sx = start; sx <= end; ++sx) {
			double nd = (w > nw) ? (sx/sf - x) : (sx - x*sf);
			double f = lanczos_tap(nd);
			if (num_coeffs == size_coeffs) {
				size_coeffs <<= 1;
				coeffs = (float *)realloc(coeffs, size_coeffs * sizeof(float));
			}
		
			coeffs[num_coeffs++] = f;
			sum += f;
		}

		for (sx = start; sx <= end; ++sx) {
			coeffs[pd[x].startcoeff + sx - start] /= sum;
		}
	}
	
	for (y = 0; y < h; ++y) {
		float *sptr = pix + y*w;
		unsigned char *dptr = npix + y*dstride;
		unsigned char ch = 0;
		for (x = 0; x < nw; ++x) {
			float acc = 0.0;
			float *cf = &coeffs[pd[x].startcoeff];
			unsigned sx;
			
			for (sx = pd[x].start; sx <= pd[x].end; ++sx) {
				acc += sptr[sx] * *cf++;
			}

			if (acc < 0.0)
				ch = 0;
			else if (acc > 255.0)
				ch = 255;
			else
				ch = (unsigned char)acc;
			*dptr++ = ch;
		}
		for ( ; x < dstride; ++x) {
			*dptr++ = ch;
		}
	}
}

void vscale(unsigned char *pix, float *npix, unsigned w, unsigned h, unsigned nh, unsigned dstride)
{
	struct pix_desc *pd = (struct pix_desc *)malloc(nh * sizeof(struct pix_desc));
	int size_coeffs = 8;
	float *coeffs = (float *)malloc(size_coeffs * sizeof(float));
	int num_coeffs = 0;
	int x, y, sy;
	double sf = (double)h / (double)nh;
	double support = (h > nh) ? (3.0 * sf) : (3.0 / sf);

	/* calculate the filter */
	for (y = 0; y < nh; ++y) {
		int start = ceil(y * sf - support);
		int end = floor(y * sf + support);
		double sum = 0.0;

		if (start < 0) {
			start = 0;
		}
		if (end > h - 1) {
			end = h - 1;
		}

		pd[y].start = start;
		pd[y].end = end;
		pd[y].startcoeff = num_coeffs;

		for (sy = start; sy <= end; ++sy) {
			double nd = (h > nh) ? (sy/sf - y) : (sy - y*sf);
			double f = lanczos_tap(nd);
			if (num_coeffs == size_coeffs) {
				size_coeffs <<= 1;
				coeffs = (float *)realloc(coeffs, size_coeffs * sizeof(float));
			}
			
			coeffs[num_coeffs++] = f;
			sum += f;
		}

		for (sy = start; sy <= end; ++sy) {
			coeffs[pd[y].startcoeff + sy - start] /= sum;
		}
	}

#if CACHE_LINE_FACTOR > 1
	for (x = 0; x < (w/CACHE_LINE_FACTOR) * CACHE_LINE_FACTOR; x += CACHE_LINE_FACTOR) {
		unsigned char *sptr = pix + x;
		float *dptr = npix + x;
		for (y = 0; y < nh; ++y) {
#if 0
			int i;
			float acc[CACHE_LINE_FACTOR];
			for (i = 0; i < CACHE_LINE_FACTOR; ++i)
				acc[i] = 0.0;
			float *cf = &coeffs[pd[y].startcoeff];
			unsigned sy;
		
			for (sy = pd[y].start; sy <= pd[y].end; ++sy) {
				for (i = 0; i < CACHE_LINE_FACTOR; ++i) {
					acc[i] += sptr[sy * w + i] * *cf;
				}
				++cf;
			}

			for (i = 0; i < CACHE_LINE_FACTOR; ++i) {
				dptr[i] = acc[i];
			}
#else
			/*
			 * xmm0 - xmm3: acc[0..15]
			 * xmm4: current filter coefficient
			 * xmm5, xmm6, xmm7: scratchpad
			 */
			asm (
				/* clear */
				"pxor %%xmm0, %%xmm0    \n"
				"pxor %%xmm1, %%xmm1    \n"
				"pxor %%xmm2, %%xmm2    \n"
				"pxor %%xmm3, %%xmm3    \n"

				/* main loop */
				".lbl:                   \n"
				
				/* a zero is useful during unpacking */
				"pxor %%xmm4, %%xmm4     \n"
				
				/* fetch all 16 source bytes */
				"movups (%0), %%xmm5     \n"
				"prefetcht0 (%0,%3,4)    \n"

				/* unpack into words (xmm5, xmm7) */
				"movaps %%xmm5, %%xmm7    \n"
				"punpcklbw %%xmm4, %%xmm5 \n"
				"punpckhbw %%xmm4, %%xmm7 \n"

				/* unpack xmm5 into dwords (xmm5, xmm6) */
				"movaps %%xmm5, %%xmm6    \n"
				"punpcklwd %%xmm4, %%xmm5 \n"
				"punpckhwd %%xmm4, %%xmm6 \n"

				/* convert xmm5, xmm6 to floats */
				"cvtdq2ps %%xmm5, %%xmm5 \n"
				"cvtdq2ps %%xmm6, %%xmm6 \n"

				/* fetch the coefficient */
				"movss (%2), %%xmm4      \n"
				"shufps $0x0, %%xmm4, %%xmm4 \n"

				/* do the muls for xmm5 and xmm6 */
				"mulps %%xmm4, %%xmm5    \n"
				"mulps %%xmm4, %%xmm6    \n"
				"addps %%xmm5, %%xmm0    \n"
				"addps %%xmm6, %%xmm1    \n"

				/* get the zero back again */
				"pxor %%xmm4, %%xmm4     \n"

				/* unpack xmm7 into dwords (xmm7, xmm6) */
				"movaps %%xmm7, %%xmm6    \n"
				"punpcklwd %%xmm4, %%xmm7 \n"
				"punpckhwd %%xmm4, %%xmm6 \n"

				/* convert xmm7, xmm6 to floats */
				"cvtdq2ps %%xmm7, %%xmm7 \n"
				"cvtdq2ps %%xmm6, %%xmm6 \n"

				/* fetch the coefficient */
				"movss (%2), %%xmm4      \n"
				"shufps $0x0, %%xmm4, %%xmm4 \n"

				/* do the second set of muls */
				"mulps %%xmm4, %%xmm7    \n"
				"mulps %%xmm4, %%xmm6    \n"
				"addps %%xmm7, %%xmm2    \n"
				"addps %%xmm6, %%xmm3    \n"

				/* move along, and loop */
				"addl $4, %2             \n"
				"addl %3, %0             \n"
				"decl %1                 \n"
				"jnz .lbl                \n"

				/* store the values */
				"movaps %%xmm0, (%4)     \n"
				"movaps %%xmm1, 16(%4)   \n"
				"movaps %%xmm2, 32(%4)   \n"
				"movaps %%xmm3, 48(%4)   \n"
				: :
				"r" (&sptr[pd[y].start * w]),        /* 0: srcptr base */
				"r" (pd[y].end - pd[y].start + 1),   /* 1: filter len */
				"r" (&coeffs[pd[y].startcoeff]),     /* 2: coeffs base */
				"r" (w),                             /* 3: stride */
				"r" (dptr)                           /* 4: dstptr base */
				: "memory", "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7"
			);
#endif
			dptr += dstride;
		}
	}
	for (x = (x/CACHE_LINE_FACTOR)*CACHE_LINE_FACTOR; x < w; ++x) {
#else
	for (x = 0; x < w; ++x) {
#endif
		unsigned char *sptr = pix + x;
		float *dptr = npix + x;
		for (y = 0; y < nh; ++y) {
			float acc = 0.0;
			float *cf = &coeffs[pd[y].startcoeff];
			unsigned sy;
			
			for (sy = pd[y].start; sy <= pd[y].end; ++sy) {
				acc += sptr[sy * w] * *cf++;
			}

			*dptr = acc;
			dptr += dstride;
		}
	}
}

int main(int argc, char **argv)
{
	unsigned nominal_w = atoi(argv[1]);
	unsigned nominal_h = atoi(argv[2]);

	unsigned samp_h0 = 2, samp_v0 = 2;
	unsigned samp_h1 = 1, samp_v1 = 1;
	unsigned samp_h2 = 1, samp_v2 = 1;
	unsigned max_samp_h = 2, max_samp_v = 2;

	unsigned nw0 = nominal_w * samp_h0 / max_samp_h, nh0 = nominal_h * samp_v0 / max_samp_v;
	unsigned nw1 = nominal_w * samp_h1 / max_samp_h, nh1 = nominal_h * samp_v1 / max_samp_v;
	unsigned nw2 = nominal_w * samp_h2 / max_samp_h, nh2 = nominal_h * samp_v2 / max_samp_v;

	unsigned stride0 = (nw0 + DCTSIZE-1) & ~(DCTSIZE-1);
	unsigned stride1 = (nw1 + DCTSIZE-1) & ~(DCTSIZE-1);
	unsigned stride2 = (nw2 + DCTSIZE-1) & ~(DCTSIZE-1);

	struct jpeg_decompress_struct dinfo;
	struct jpeg_error_mgr jerr;
	dinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&dinfo);
	jpeg_stdio_src(&dinfo, stdin);
	jpeg_read_header(&dinfo, TRUE);
	dinfo.raw_data_out = TRUE;
	jpeg_start_decompress(&dinfo);

	fprintf(stderr, "Scaling using Lanczos filter:\n");
	fprintf(stderr, "  Y component: %ux%u -> %ux%u\n", dinfo.comp_info[0].width_in_blocks * DCTSIZE, dinfo.comp_info[0].height_in_blocks * DCTSIZE, nw0, nh0);
	fprintf(stderr, "  Cb component: %ux%u -> %ux%u\n", dinfo.comp_info[1].width_in_blocks * DCTSIZE, dinfo.comp_info[1].height_in_blocks * DCTSIZE, nw1, nh1);
	fprintf(stderr, "  Cr component: %ux%u -> %ux%u\n", dinfo.comp_info[2].width_in_blocks * DCTSIZE, dinfo.comp_info[2].height_in_blocks * DCTSIZE, nw2, nh2);

	JSAMPLE *data_y  = (JSAMPLE*)memalign(16, dinfo.comp_info[0].height_in_blocks * dinfo.comp_info[0].width_in_blocks * DCTSIZE * DCTSIZE);
	JSAMPLE *data_cb = (JSAMPLE*)memalign(16, dinfo.comp_info[1].height_in_blocks * dinfo.comp_info[1].width_in_blocks * DCTSIZE * DCTSIZE);
	JSAMPLE *data_cr = (JSAMPLE*)memalign(16, dinfo.comp_info[2].height_in_blocks * dinfo.comp_info[2].width_in_blocks * DCTSIZE * DCTSIZE);
	JSAMPLE *data_ny, *data_ncb, *data_ncr;

	int total_lines = 0, blocks = 0;
	while (total_lines < dinfo.comp_info[0].height_in_blocks * DCTSIZE) {
		unsigned max_lines = dinfo.max_v_samp_factor * DCTSIZE;

		JSAMPROW y_row_ptrs[max_lines];
		JSAMPROW cb_row_ptrs[max_lines];
		JSAMPROW cr_row_ptrs[max_lines];
		JSAMPROW* ptrs[] = { y_row_ptrs, cb_row_ptrs, cr_row_ptrs };
		int i;

		for (i = 0; i < max_lines; ++i) {
			y_row_ptrs[i]  = data_y  + (i+blocks*DCTSIZE*dinfo.comp_info[0].v_samp_factor) * dinfo.comp_info[0].width_in_blocks * DCTSIZE;
			cb_row_ptrs[i] = data_cb + (i+blocks*DCTSIZE*dinfo.comp_info[1].v_samp_factor) * dinfo.comp_info[1].width_in_blocks * DCTSIZE;
			cr_row_ptrs[i] = data_cr + (i+blocks*DCTSIZE*dinfo.comp_info[2].v_samp_factor) * dinfo.comp_info[2].width_in_blocks * DCTSIZE;
		}
		
		total_lines += max_lines;
		++blocks;

		if (jpeg_read_raw_data(&dinfo, ptrs, max_lines) == 0)
			break;
	}

	{
		float *npix = (float*)memalign(16, dinfo.comp_info[0].width_in_blocks * DCTSIZE * nh0 * sizeof(float));	
		vscale(data_y, npix, dinfo.comp_info[0].width_in_blocks * DCTSIZE, dinfo.comp_info[0].height_in_blocks * DCTSIZE, nh0, dinfo.comp_info[0].width_in_blocks * DCTSIZE);
		data_ny = (unsigned char *)malloc(nw0 * stride0);
		hscale(npix, data_ny, dinfo.comp_info[0].width_in_blocks * DCTSIZE, nh0, nw0, stride0);
		free(npix);
	}
	{
		float *npix = (float*)memalign(16, dinfo.comp_info[1].width_in_blocks * DCTSIZE * nh1 * sizeof(float));	
		vscale(data_cr, npix, dinfo.comp_info[1].width_in_blocks * DCTSIZE, dinfo.comp_info[1].height_in_blocks * DCTSIZE, nh1, dinfo.comp_info[1].width_in_blocks * DCTSIZE);
		data_ncr = (unsigned char *)malloc(nw1 * stride1);
		hscale(npix, data_ncr, dinfo.comp_info[1].width_in_blocks * DCTSIZE, nh1, nw1, stride1);
		free(npix);
	}
	{
		float *npix = (float*)memalign(16, dinfo.comp_info[2].width_in_blocks * DCTSIZE * nh2 * sizeof(float));	
		vscale(data_cb, npix, dinfo.comp_info[2].width_in_blocks * DCTSIZE, dinfo.comp_info[2].height_in_blocks * DCTSIZE, nh2, dinfo.comp_info[2].width_in_blocks * DCTSIZE);
		data_ncb = (unsigned char *)malloc(nw2 * stride2);
		hscale(npix, data_ncb, dinfo.comp_info[2].width_in_blocks * DCTSIZE, nh2, nw2, stride2);
		free(npix);
	}
	jpeg_destroy_decompress(&dinfo);
	
	struct jpeg_compress_struct cinfo;
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);
	jpeg_stdio_dest(&cinfo, stdout);
	cinfo.input_components = 3;
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, 85, FALSE);
	cinfo.image_width = nominal_w;
	cinfo.image_height = nominal_h;
	cinfo.raw_data_in = TRUE;
	jpeg_set_colorspace(&cinfo, JCS_YCbCr);
	cinfo.comp_info[0].h_samp_factor = samp_h0;
	cinfo.comp_info[0].v_samp_factor = samp_v0;
	cinfo.comp_info[1].h_samp_factor = samp_h1;
	cinfo.comp_info[1].v_samp_factor = samp_v1;
	cinfo.comp_info[2].h_samp_factor = samp_h2;
	cinfo.comp_info[2].v_samp_factor = samp_v2;
	jpeg_start_compress(&cinfo, TRUE);

	total_lines = 0;
	blocks = 0;
	while (total_lines < cinfo.comp_info[0].height_in_blocks * DCTSIZE) {
		unsigned max_lines = cinfo.max_v_samp_factor * DCTSIZE;

		JSAMPROW y_row_ptrs[max_lines];
		JSAMPROW cb_row_ptrs[max_lines];
		JSAMPROW cr_row_ptrs[max_lines];
		JSAMPROW* ptrs[] = { y_row_ptrs, cb_row_ptrs, cr_row_ptrs };
		int i;

		for (i = 0; i < max_lines; ++i) {
			// simple edge extension
			int yline = i + blocks*DCTSIZE*cinfo.comp_info[0].v_samp_factor;
			if (yline > nh0 - 1)
				yline = nh0 - 1;

			int cbline = i + blocks*DCTSIZE*cinfo.comp_info[1].v_samp_factor;
			if (cbline > nh1 - 1)
				cbline = nh1 - 1;

			int crline = i + blocks*DCTSIZE*cinfo.comp_info[2].v_samp_factor;
			if (crline > nh2 - 1)
				crline = nh2 - 1;

			y_row_ptrs[i]  = data_ny  + yline * stride0;
			cb_row_ptrs[i] = data_ncb + cbline * stride1;
			cr_row_ptrs[i] = data_ncr + crline * stride2;
		}
		
		total_lines += max_lines;
		++blocks;

		jpeg_write_raw_data(&cinfo, ptrs, max_lines);
	}
	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);

	return 0;
}

